--TODO LATER: SetFrameLevel seems to be slightly nondeterministic
	-- probably because we're trying to overlap 4 other polys with the same framelevel
	-- we fixed it, maybe, but we're using about 125 framelevels across the entire worldgrid.
	-- can we make it more efficient?

--TODO: redetermine min/max billboard display angle range

--TODO: consider changing the color of the diffuse light sources themselves

--TODO: rerender.  even if we don't rerender, recopy to psd -- bumpers probably messed up by merging onto stroke effect
	-- still not sure whether or not we want to use inner stroke effect regardless	
	--TODO: maybe apply stroke effect to rings?


-------------------------------------------------------------------------------------------
-------------------------------------------------------------------------------------------
--
-- Checkerboard grid:
-- We start with flat 8x8 grid and pinch the corners down to resemble a sphere.
-- The sphere is then rotated and stretched in accordance with the camera perspective.
-- To the best of our very limited knowledge, the mathematics involved are "real"
-- and could be replicated 1:1 in any 3D rendering system.
-- 
-- Spheres and rings:
-- This gets weird.  Spheres don't use the same perspective as tiles.
-- In particular, spheres shrink very quickly into the horizon when far from the player,
-- and they rotate very quickly toward the camera when close to the player.
-- We begin by calculating the sphere's "real" position on the grid.
-- From there, we take the sphere's real Z value and apply various scaling to it
-- to get the sphere's "fake" size (which is then used to recalculate its position).
-- The sphere's displayed rotation is a combination of the slope of the sphere at its vertex
-- AND the angle between the camera and the sphere,
-- though we add additional "fake" distance to the camera for a shallower angle.
-- The exact angle, of course, is an arbitrary number chosen to look good
-- (the prerendered spheres are rendered under equally arbitrary settings).
-- 
-- Player position:
-- 99% smoke and mirrors.  The player modelframe is not affected by camera,
-- though some of the hardcoded numbers are supposed to reflect what the camera transforms would be.
--
-------------------------------------------------------------------------------------------
-------------------------------------------------------------------------------------------


--TODO: fast graphics mode doesn't seem to be much of an improvement anymore.
	-- maybe because we've added 225 billboarded spheres since then?
	-- make sure that the unused tile polys are actually going unused
	-- TODO: also cull some billboards

--TODO LATER: there's probably a better solution to removing seams than doubling up on polys


do
	local super=TRAININGGROUNDS_Drawable
	TRAININGGROUNDS_SECRET35_Drawable_WorldGrid = TRAININGGROUNDS_inheritsFrom(super)
	local class=TRAININGGROUNDS_SECRET35_Drawable_WorldGrid
	
	-- Size of the base frame which contains the rest of the world drawable
	local width=100
	local height=width
	--this is tricky:
	-- polygons are locked to TOPLEFT anchor point
	-- but we want the grid to rotate around parentframe's CENTER
	-- so we need to manually offset the polygons by size/2.
	-- also note that we add offset to x, but subtract it from y
	local polyoffsetx=width/2		
	local polyoffsety=polyoffsetx
	-- we also shift the world vertically to better line up with the original game
	--(screenoffsety was moved to camera)
	

	
	
	-- returns SX, SZ, SY
	-- optional arg cangz (height axis, horizontal camera turn).
	-- if cangz is specified, function returns a 4th arg, slope_angle
		-- (angle of the sphere surface, used for billboard display angle.)
		-- remember to subtract cangx (cameraanglex) after this function ends to get the proper display angle!
		-- also remember that CalcSphereCurve's angle changes linearly with distance on the sphere.
			-- to pull off the perspective trick in the original game, you'll need to combine it with something else.		
	local function CalcSphereCurve(x,z,height,cangz)		
		local slope_angle=nil
		-- args: x and z are distance from the "top center" of the sphere
		-- and height is the distance *ABOVE the SURFACE* of the sphere
		-- 		(sphere radius is hardcoded) (currently in Camera, might get moved back later)
		--		(so height will be 0 for tiles and billboard_default_height for spheres)

		-- The sphere (is intended to) curve 90 degrees downwards at a distance of sphereradius.
			-- Total circumference of the sphere = 2*PI*radius
			-- Divide that by 4 to get quartercircumference
		local spherequartercircumference=(2*math.pi*sphereradius)/4		
		local xs=x/sphereradius
		local zs=z/sphereradius
	
		-- first, map x/z square to circle
		--SOURCE: stackoverflow.com/questions/1621831/how-can-i-convert-coordinates-on-a-square-to-coordinates-on-a-circle
		local xc=x*math.sqrt(1-0.5*zs*zs)
		local zc=z*math.sqrt(1-0.5*xs*xs)
	
		-- we now have cartesian coordinates for a circle.  convert to polar coordinates.
		local s=math.atan2(zc,xc)
		local dist=math.sqrt(xc*xc+zc*zc)
		-- the distance coordinate becomes our phi angle for the sphere
		local t=math.pi/2*(dist/spherequartercircumference)
		
		if(cangz)then
			-- for billboard reasons, we also want to save the angle of the slope at the current point
			-- ...BUT we want to treat the angle as though it's measured from an arbitrary distance behind us.
			-- say, 45 degrees' worth of distance away?  to match the camera angle?
			-- we also need to rotate the point around the camera now.
			--TODO LATER: this seems wasteful, is it possible to reduce the number of rotation transforms?
			local xc2=xc*math.cos(cangz) - zc*math.sin(cangz)
			local zc2=zc*math.cos(cangz) + xc*math.sin(cangz)			
			local arbitrary_angle=math.pi/4				
			local arbitrary_distance=(arbitrary_angle*2/math.pi)*spherequartercircumference		
			local zc3=zc2+arbitrary_distance
			local dist2=math.sqrt(xc2*xc2+zc3*zc3)
			local t2=math.pi/2*(dist2/spherequartercircumference)
			-- ...then subtract arbitary_angle.
			-- the intended result is that when directly in front of the player, 
			-- the angle is the same as if it had been measured from 0,0
			-- while behind the player, the angle is negative (instead of positive as radial coordinates would work from 0,0)
			-- meanwhile, to the sides, slope_angle still increases slightly 
			slope_angle=t2-arbitrary_angle
		end
		
		-- we want to find the point on the edge of the sphere
			-- with a radius of sphereradius
			-- and, if specified, extra height above (or below) the sphere.
		local r=sphereradius+height
		
		
		--SOURCE: stackoverflow.com/questions/969798/plotting-a-point-on-the-edge-of-a-sphere
		local sx,sy,sz		
		
		sx=r*math.cos(s)*math.sin(t)
		sz=r*math.sin(s)*math.sin(t)
		sy=r*math.cos(t)
		
		-- this should give us the point on the sphere.
		-- however we're expecting sz==0 to be the *top* of the sphere, not the center
		-- so we offset the results by the initial sphereradius.
		sy=sy-sphereradius
		
		-- remember: camera angle has not yet been accounted for;
				-- "reverse sphere" angle has not yet been acounted for.
				-- deal with these after exiting this function
		
		
		return sx,sz,sy,slope_angle
		
	end
	
	
	
	-- Width of the grid in TILES.  Should be an EVEN number.
	-- There will be GRID_WIDTH+1 billboards, drawn on the vertices of the tiles.
	-- Any time you see a check for x<=GRID_WIDTH, we're in a GRID_WIDTH+1 loop and we're checking for tiles only.
	local GRID_WIDTH=14
	
	local POLYS_PER_TILE=4
	--local POLYS_PER_TILE=2 -- "fast graphics mode"
	

	------------------------------------------------------------------------------------------
	-- syntax: subframe:SetSubframeVertices
	local function SetSubframeVertices(subframe,x1,y1,x2,y2,x3,y3,bottomclip)
		-- no need to render polys which are below the bottom of the screen...
		bottomclip=bottomclip-polyoffsety
		--TODO LATER: also clip polys off the left or right sides
		local ystretch=1		
		if(y1>=bottomclip or y2>=bottomclip or y3>=bottomclip)then
			subframe:Show()		
			TAXIFRAME_TextureAdd(subframe,100,1,1,1,x1*0.01,-y1*0.01,x2*0.01,-y2*0.01,x3*0.01,-y3*0.01)		
		else
			subframe:Hide()
		end
	end
	
	function SetSubframeColor(subframe,R,G,B,A)
		subframe.texture:SetGradientAlpha("HORIZONTAL",R,G,B,A,R,G,B,A)
	end	
	
	function class:QuickSubframeSetup()
		local subframe=TRAININGGROUNDS_ReusableFrames:GetFrame()				
		subframe:Show()
		subframe:SetParent(self.frame)
		subframe:SetAllPoints(self.frame)			
		subframe.texture:Show()		
		subframe:SetAlpha(1)
		tinsert(self.allsubframes,subframe)	
		return subframe
	end
	
	function class:QuickBillboardSetup(x,y)
		local billboard=TRAININGGROUNDS_ReusableFrames:GetFrame()				
		billboard:Show()
		billboard:SetParent(self.frame)
		billboard:SetAllPoints(self.frame)		
		billboard.texture:Show()		
		billboard.texture:SetColorTexture(0.5,0.5,1,1)		
		if(x==8 and y==8)then
			billboard.texture:SetTexture("Interface\\AddOns\\TrainingGrounds\\Assets\\Graphics\\Misc\\circle.tga")
		else
			billboard.texture:SetTexture("Interface\\AddOns\\TrainingGrounds_GetSpheres\\Assets\\Graphics\\rendertest_springs.tga")
			local p=1/1024.0;
			billboard.texture:SetTexCoord(p*0,p*128,p*0,p*128)
		end
		billboard:SetAlpha(1)
		tinsert(self.allbillboards,billboard)	
		return billboard
	end
	
	function class:SetBillboardAppearance(x,y,gx,gy,currenttime)		
		--x,y refers to the array index
		--gx,gy refers to the position in the world (outside of this module, we'd say "which vertex?")
				
		--TODO NEXT: sphere texture picker needs currenttime passed to it
		local sphere=self.owner:GetSphereAt(gx,gy)
		
		local filename,offset,index,isaring,isasparkle
		if(sphere:IsANil())then
			self.billboardgrid[x][y]:Hide()
			return
		else
			self.billboardgrid[x][y]:Show()			
			--TODO NEXT: pass in currenttime ... from somewhere
			data=sphere:GetDrawData(currenttime,self.owner.environment.game.scenario)
			filename=data.file
			offset=data.offset			
			isaring=sphere:IsARing()
			isasparkle=sphere:IsASparkle()
		end
		
		local billboard=self.billboardgrid[x][y]
		local angle=self.billboard_display_angle[x][y]
		--TODO NEXT: recalculate angle range
		local min_angle=-75/180.0*math.pi
		local max_angle =5/180.0*math.pi
		if(isaring)then
			angle_index=sphere:GetRingIndex(currenttime)
		elseif(isasparkle)then
			angle_index=sphere:GetRingIndex(currenttime)
		else
			angle_index=33-(1+math.floor(32*(angle-min_angle)/(max_angle-min_angle)))
		end
		if(angle_index<1)then angle_index=1 end
		if(angle_index>32)then angle_index=32 end		
		--billboard.texture:SetTexture("Interface\\AddOns\\TrainingGrounds_GetSpheres\\Assets\\Graphics\\rendertest_springs.tga")
		billboard.texture:SetTexture(filename)
		local p=1/1024.0;
		local tx=(angle_index-1) % 8
		local ty=math.floor((angle_index-1) / 8)
		 -- if(x==8)then
			 -- print("SBT",x,y,angle*180/math.pi,angle_index)
		 -- end
		tx=tx*128
		ty=ty*128+offset*(4*128)
		billboard.texture:SetTexCoord((tx+0)*p,(tx+128)*p,(ty+0)*p,(ty+128)*p)
		-- local f=1
		-- if(sphere.flashendtime)then
			-- local timediff=sphere.flashendtime-self.environment.localtime
			-- f=1-(timediff*1.5)
			-- if(timediff<0)then
				-- sphere.flashendtime=nil
			-- end
		-- end
		-- billboard.texture:SetGradientAlpha("HORIZONTAL",f,f,f,1,f,f,f,1)
		
		
	end
	
	
	
	function class:SetCustomInfo()
		super.SetCustomInfo(self)
		self.anchor="CENTER"
		self.allsubframes={}
		self.subframegrid={}
		self.allbillboards={}
		self.billboardgrid={}
	end
	
	function class:CreateFrame()
		self.frame=TRAININGGROUNDS_ReusableFrames:GetFrame()
		self.frame:Show()
	end


	
	--NOTE: colors will display differently based on System menu brightness/contrast settings.
	--TODO: color changes depending on stage
	--local color1={238/256.0,136/256.0, 0/256,1.0}
	--local color2={102/256.0, 34/256.0, 0/256,1.0}
	
		

		
	
	function class:SetFrameAppearance()	
		self.width=width
		self.height=height
		self.frame:SetParent(self.parentframe)
		self.frame:SetPoint("CENTER",self.parentframe,"CENTER",0,0)
		self.frame:Show()		
		self.frame:SetWidth(self.width)
		self.frame:SetHeight(self.height)	
		
		local color1=self.owner.color1
		local color2=self.owner.color2

		-- for each square tile on the grid,
		-- create the four subframes that make up that tile
			-- (TODO: in fast graphics mode, skip creating the 3rd+4th polys)
		-- and create a billboard frame at the vertex.
		-- we have not yet assigned positions to these frames
			-- (that comes later, in ArrangeFrame).
		for x=1,GRID_WIDTH+1 do
			self.subframegrid[x]={}
			self.billboardgrid[x]={}
			for y=1,GRID_WIDTH+1 do
				
				if(x<=GRID_WIDTH and y<=GRID_WIDTH)then
					self.subframegrid[x][y]={}
					
					local subframe1,subframe2,billboard1
					subframe1=self:QuickSubframeSetup()
					self.subframegrid[x][y][1]=subframe1
									
					subframe2=self:QuickSubframeSetup()
					self.subframegrid[x][y][2]=subframe2
					
					subframe3=self:QuickSubframeSetup()
					self.subframegrid[x][y][3]=subframe3

					subframe4=self:QuickSubframeSetup()
					self.subframegrid[x][y][4]=subframe4
					
					local intx,inty
					intx=math.floor((x-1))	-- we used to divide by grid subdivisions, before we removed grid subdivisions
					inty=math.floor((y-1))
					if((intx%2) ~= (inty%2))then
						SetSubframeColor(subframe1,unpack(color1))
						SetSubframeColor(subframe2,unpack(color1))
						SetSubframeColor(subframe3,unpack(color1))
						SetSubframeColor(subframe4,unpack(color1))
					else
						SetSubframeColor(subframe1,unpack(color2))
						SetSubframeColor(subframe2,unpack(color2))
						SetSubframeColor(subframe3,unpack(color2))
						SetSubframeColor(subframe4,unpack(color2))
					end
				end
				
				billboard1=self:QuickBillboardSetup(x,y)
				self.billboardgrid[x][y]=billboard1
			end
		end		
		
		-- keep track of vertices separately to avoid recalculating them 4x for each tile cornering them
		self.gridverts={}	-- tile corners
		self.billboardverts={}	-- base for billboarded sphere sprites.  coord set to nil if object is behind-ish camera.
		self.billboard_size={}	
		self.billboard_in_range={}	-- billboards are only drawn within 6.125 units hz or vt
		self.billboard_distance={}	-- beyond a certain distance, billboards appear behind the grid
										-- TODO: probably no longer used
		self.billboard_display_angle={}	-- angle at which the ground is sloping away at a given point (relative to camera)
		
		
		for x=1,GRID_WIDTH+1 do
			self.gridverts[x]={}
			self.billboardverts[x]={}
			self.billboard_size[x]={}
			self.billboard_in_range[x]={}
			self.billboard_distance[x]={}	
			self.billboard_display_angle[x]={}			
			for y=1,GRID_WIDTH+1 do
				self.gridverts[x][y]={}	
			end
		end
		
	end
	

	
	function class:SetFrameLevel()
		local tilemid=(GRID_WIDTH+1)/2	--7.5
		local bbmid=(GRID_WIDTH)/2+1	--8
		--14x14 tiles, 15x15 billboards
		for x=1,GRID_WIDTH+1 do
			for y=1,GRID_WIDTH+1 do
				local tilexdist=math.abs(x-tilemid)-0.5
				local tileydist=math.abs(y-tilemid)-0.5
				
				local bbxdist=math.abs(x-bbmid)
				local bbydist=math.abs(y-bbmid)				
				
				local tiletaxidist=tilexdist+tileydist
				local bbtaxidist=bbxdist+bbydist
				-- adjust billboards directly in front of the player so they're drawn behind the horizon
					--TODO LATER: double-check numbers, make sure this actually works
					--also, see if we want to adjust ALL billboards on the sphere (or if we actually can't do that due to off-by-one reasons)
					--also, maybe we only need to adjust the very farthest billboards instead of all of them in a line?
				if(bbxdist==0 or bbydist==0)then
					bbtaxidist=bbtaxidist+1
				end

				
				
				--local tileframelevel=self.framelayer.level+(GRID_WIDTH*2-tiletaxidist)
				local tilemultiplier=4
				local tileframelevel=self.framelayer.level+(GRID_WIDTH*tilemultiplier*2-tiletaxidist*tilemultiplier)
				if(tileframelevel<1)then tileframelevel=1 end	--TODO: assertion failed
				if(x<=GRID_WIDTH and y<=GRID_WIDTH)then
					for i=1,POLYS_PER_TILE do					
						self.subframegrid[x][y][i]:SetFrameLevel(tileframelevel)
					end
				end
				--dubious:				
				--local billboardframeleveloffset=2
				local billboardframeleveloffset=1+8
				--local billboardframelevel=tileframelevel+5
				local billboardframelevel=billboardframeleveloffset+self.framelayer.level+(GRID_WIDTH*tilemultiplier*2-bbtaxidist*tilemultiplier)
				--(DEPRECATED) (maybe) (-ish)				
				-- SetFrameLevel may be called before ArrangeFrame, so we need to check for nil				
				-- if(bbtaxidist>MAX_BILLBOARD_GROUNDCLIP_DISTANCE)then				
					-- ----TODO: make sure this can't bring framelevel below 1
					-- billboardframelevel=billboardframelevel-10
				-- end
					
				----if(math.max(xdist,ydist)>5)then billboardframelevel=billboardframelevel-10 end
				self.billboardgrid[x][y]:SetFrameLevel(billboardframelevel)
				
				  -- if((x==8 and y==8) or (false))then
					  -- print("Tileframelevel at",x,y,":",tileframelevel,tilexdist,tiletaxidist,tilemid)
					  -- print("BBframelevel at",x,y,":",billboardframelevel,bbtaxidist)
				  -- end				
			end
		end
	end	
	
	
	
	
	function class:SetFrameLayer()
		--override; don't call super while overriding
		self.framelayer=self.environment.cameramanager.arenaborderframelayer
	end
	
	--TODO: rename some of these
	--cposx/cposy: position of the CHARACTER
	--cposz: position of the CAMERA
	--coffsetx/coffsety: supposed to be camera offset, but not really implemented correctly.  don't use.
	function class:ArrangeFrame(parentanchor,cposx,cposy,cposz,cangx,cangy,cangz,coffsetx,coffsety)
		self.frame:SetParent(self.parentframe)
		if(self.parentframe==UIParent)then		
			print("WARNING:",self.frame.name,"parent is UIParent")
		end
		if(self.frame:GetParent()==UIParent)then
			print("WARNING:",self.frame.name,"parent is UIParent")
		end
		self.frame:ClearAllPoints()
		self.frame:SetPoint(self.anchor,self.parentframe,parentanchor,0,0)
		self:SetFrameSizeAndOrientation(100,100,0,0,0)
	
		local az=cangz--GetTime()*0.5*-1		-- "height axis", used to turn horizontally
		local ax=cangx
		local ay=cangy		
		local cz_pos=cposz		--"negative z" is vertical up.
		local cx_pos=cposx
		local cy_pos=cposy

		
		-- if we simply round to floor, then cx_pos/cy_pos will always be positive.
		-- we want pos variables to range from positive to negative
		-- or else grid visibility will change by 1 tile depending on which direction we're facing.
		--local gridx=math.floor(cposx)
		--local gridy=math.floor(cposy)
		local gridx=math.floor(cposx+0.5)
		local gridy=math.floor(cposy+0.5)
		
		cx_pos=cposx-gridx
		cy_pos=cposy-gridy		
				
		local parity=((gridx%2)~=(gridy%2))
		for x=1,GRID_WIDTH do
			for y=1,GRID_WIDTH do
				intx=math.floor((x-1))	--we used to divide by grid subdivisions here
				inty=math.floor((y-1))
				local tileparity=((intx%2)~=(inty%2))
				local color
				local color1=self.owner.color1
				local color2=self.owner.color2
				if(tileparity==parity)then
					tilecolor=color1
				else
					tilecolor=color2
				end
				for i=1,POLYS_PER_TILE do
					SetSubframeColor(self.subframegrid[x][y][i],unpack(tilecolor))
				end
			end
		end
		
		for y=1,GRID_WIDTH+1 do
			local x8_slope_angle
			-- "for i=1 to 15, but start from 8 and wrap around and end at 7"
			for fake_x=1,GRID_WIDTH+1 do
		--		123456789012345		<- we started with these numbers			
				x=((fake_x+6)%(GRID_WIDTH+1))+1					
		--		890123451234567		<- but now we have these numbers.
	-- (this trick was leftover from a time when we planned to use the "middle" column for further calculations)
		
				local xx,yy,zz,curveheight,slope_angle
				--TODO: rename "old" vars to something like "relative_to_center" (but pick a much better name)
						-- ...relx?
				local oldxx,oldyy,oldzz
				-- calculate tile position relative to "center" of grid
				oldxx=(-1-GRID_WIDTH/2)
				oldyy=(-1-GRID_WIDTH/2)
				oldxx=oldxx+x
				oldyy=oldyy+y
				oldxx=oldxx-0
				oldyy=oldyy-0
				oldxx=oldxx-cx_pos
				oldyy=oldyy-cy_pos
						

					
				--TODO: cull billboards beyond sides/below screen, if possible					
				-- original game has a draw distance of 6.125.  for the moment, we're ignoring draw distance.
				--if(math.abs(oldxx)<=6.125 and math.abs(oldyy)<=6.125)then				
				if(true)then
					self.billboard_in_range[x][y]=true
				else					
					self.billboard_in_range[x][y]=false
				end
				-- TODO: we might be able to skip sqrt and compare to sqr_max_distance
				self.billboard_distance[x][y]=math.sqrt(oldxx*oldxx+oldyy*oldyy)
				
		
				
				-- bend the grid downwards so it resembles a sphere
				-- (slope_angle will be discarded)
				local rotated_xx,rotated_yy
				
				xx,yy,curveheight=CalcSphereCurve(oldxx,oldyy,0)
				
				-- if((x==8 and y==8))then
					-- print("cy_pos at",x,y,":",cy_pos)
				-- end		
				
				
				--TODO: WARNING: "y" now refers to lateral grid position, not height
				-- (whereas in CalcSphereCurve it was switched with "z")
				zz=-curveheight
				
				-- at this point we can add in camera offsets (if we're still using them?  i don't think x/y are used...)
				xx=xx+coffsetx
				yy=yy+coffsety
				zz=zz-cz_pos				

				-- Now that we have "spherecurve" coordinates, convert them to window coordinates 								
				
				-- Tiles first:				
				do
					local x1,y1,z1
					-- Rotate the sphere according to camera angle
					x1,y1,z1=TRAININGGROUNDS_SECRET35_MatrixMultiplicationRotation(xx,yy,zz,ax,ay,az)

					-- Stretch the view according to distance and perspective
					x1,y1=TRAININGGROUNDS_SECRET35_PerspectiveTransform(x1,y1,z1,far)
					-- We should now have a rendered image of the checkerboard grid.
					-- Center it on the window and stretch it to fit:
					x1=x1*WORLD_SCALE_MULTIPLIER
					y1=y1*WORLD_SCALE_MULTIPLIER					
					-- tile_ vars are a bit redundant, but leftover from an earlier experiment
						--TODO: remove tile_ vars and re-rename the vars that point at them
					local tile_x1=x1+polyoffsetx
					local tile_y1=y1-polyoffsety-screenoffsety
					
					--if(x==8 and y==8)then
						--print("Draw_World35 ArrangeFrame",tile_x1,tile_y1)
					--end
					
					self.gridverts[x][y]={tile_x1,tile_y1}	--add offset to x, subtract offset from y
				end
				
				
				-- Billboards:
				do
					-- start by calculating "real" billboard position					
					local billboard_default_height=0.1
					local bbxx,bbyy,bbzz,slope_angle=CalcSphereCurve(oldxx,oldyy,billboard_default_height,az)
					bbzz=-bbzz
					bbxx=bbxx+coffsetx
					bbyy=bbyy+coffsety
					bbzz=bbzz-cz_pos		

					
					local x1,y1,z1				
					x1,y1,z1=TRAININGGROUNDS_SECRET35_MatrixMultiplicationRotation(bbxx,bbyy,bbzz,ax,ay,az)
					local oldx1=x1
					local oldy1=y1								
					local oldz1=z1					

					
					----There is definitely trickery involved in the sizing of billboards.
					--Examination of the original game reveals that billboards shrink into the horizon
					--much faster than the checkerboard grid shrinks into the horizon.
								
					--current method: use same perspective distortion, but increase Z first.
					--	("diameter" has an arbitary unit of measurement.
					--   long ago, it was approximately equal to percentage of a tile,
					--   but that no longer applies)
					local default_billboard_diameter=.325
					if(self.owner.doublesize)then
						default_billboard_diameter=default_billboard_diameter*2
					end
					local default_billboard_radius=default_billboard_diameter*0.5
					local cheat_z1=oldz1
					if(cheat_z1>2)then
						cheat_z1=cheat_z1+math.pow(cheat_z1-2,1.5)
					end

					
					local cheat_x1
					cheat_x1=TRAININGGROUNDS_SECRET35_PerspectiveTransform(oldx1,oldy1,cheat_z1,far)					
					cheat_x1=cheat_x1*WORLD_SCALE_MULTIPLIER
					local x2=TRAININGGROUNDS_SECRET35_PerspectiveTransform(oldx1+default_billboard_radius,oldy1,cheat_z1,far)
					x2=x2*WORLD_SCALE_MULTIPLIER
					local displayed_billboard_radius=math.abs(x2-cheat_x1)					
					self.billboard_size[x][y]=(displayed_billboard_radius)*2
					
					
					-- we've already rotated the world in accordance with camera angle
					-- now measure the angle between sphere and camera (remember, camera angle is effectively 0 post-rotation)
					-- and calculate the difference between that and slope_angle					
					-- reminder: "y" is height, "z" is depth
					local da_z1=oldz1
					da_z1=da_z1*1+3		-- arbitary additional camera distance	
					da_z1=math.sqrt(da_z1*da_z1+oldx1*oldx1)
					if(da_z1<0)then da_z1=0 end
					local camera_to_sphere_angle=math.atan2(oldy1,da_z1)
					local rotated_slope_angle=slope_angle-cangx
					local display_angle=camera_to_sphere_angle+rotated_slope_angle
					self.billboard_display_angle[x][y]=display_angle					

					--multiplier is arbitarily chosen and must be recalculated if billboard_radius ever changes
					local billboard_offset_radius_multiplier=0.003
					--NOTE: "oldxx" and "oldyy" are not the same as "oldx1" and "oldy1" -- really need to rename these
					
					local bbxx,bbyy,bbzz=CalcSphereCurve(oldxx,oldyy,displayed_billboard_radius*billboard_offset_radius_multiplier+self.owner.billboard_height)

					--print("bbzz:",bbzz)

					bbzz=-bbzz
					bbxx=bbxx+coffsetx
					bbyy=bbyy+coffsety
					bbzz=bbzz-cz_pos									
					local x1,y1,z1				
					x1,y1,z1=TRAININGGROUNDS_SECRET35_MatrixMultiplicationRotation(bbxx,bbyy,bbzz,ax,ay,az)
					local oldx1=x1
					local oldy1=y1								
					local oldz1=z1					
					if(oldz1>-1)then 
						bbx1,bby1=TRAININGGROUNDS_SECRET35_PerspectiveTransform(oldx1,oldy1,oldz1,far)
						bbx1=bbx1*WORLD_SCALE_MULTIPLIER
						bby1=bby1*WORLD_SCALE_MULTIPLIER												
						self.billboardverts[x][y]={bbx1+polyoffsetx,bby1-polyoffsety-screenoffsety}	
					else
						self.billboardverts[x][y]=nil
					end
				end
					
			end
		end

		local bottomclip=-self.parentframe:GetHeight()/2
		
		for x=1,GRID_WIDTH+1 do
			for y=1,GRID_WIDTH+1 do

				local x1,y1,z1,x2,y2,z2,x3,y3,z3,x4,y4,z4,bbx,bby
				bbx,bby=nil,nil
				if(self.billboardverts[x][y]~=nil)then
					bbx,bby=unpack(self.billboardverts[x][y])
				end
				if(x<=GRID_WIDTH and y<=GRID_WIDTH)then
					x1,y1=unpack(self.gridverts[x][y])
					x2,y2=unpack(self.gridverts[x+1][y])
					x3,y3=unpack(self.gridverts[x+1][y+1])
					x4,y4=unpack(self.gridverts[x][y+1])
					
					
					-- TODO LATER: we want to somehow overlap the triangles very slightly to avoid showing seams along the diagonal
						-- (then we can use two subframes instead of four)
					
					--if(y1>0 or y2>0 or y3>0 or y4>0)then 
					if(true)then
						SetSubframeVertices(self.subframegrid[x][y][1],x1,y1,x2,y2,x4,y4,bottomclip)
						SetSubframeVertices(self.subframegrid[x][y][2],x3,y3,x4,y4,x2,y2,bottomclip)
						if(POLYS_PER_TILE>2)then
							SetSubframeVertices(self.subframegrid[x][y][3],x2,y2,x3,y3,x1,y1,bottomclip)
							SetSubframeVertices(self.subframegrid[x][y][4],x4,y4,x1,y1,x3,y3,bottomclip)				
						end
					end
				end
				self.billboardgrid[x][y]:ClearAllPoints()
				-- for some reason, we applied polyoffset to billboard coordinates earlier
					-- so we need to undo that now
					--TODO: find out if it's possible to NOT apply polyoffset earlier
				if(bbx and bby)then
					self.billboardgrid[x][y]:SetPoint("CENTER",self.frame,"CENTER",bbx-polyoffsetx,bby+polyoffsety)
				end
				
				
				local size=self.billboard_size[x][y]				
				-- the upper bounds on size check was supposed to prevent some weird things from occurring when billboards are behind the camera
				
				--if(x==8 and y==8)then
				if(self.billboard_in_range[x][y] and size>1 and size<170 and bbx and bby)then
					self.billboardgrid[x][y]:SetSize(size,size)
					--show/hide has been moved to SetBillboardAppearance
					--TODO: ideally move this range/size check there too.
					-- will probably involve passing function args.
					
					--self.billboardgrid[x][y]:Show()
					--TODO: rename gridx/gridy + gx/gy 
					--TODO LATER: midpoint not tested to work with widths other than 14
					local midpoint=GRID_WIDTH/2+1
					local gx=(x-midpoint)+gridx
					local gy=(y-midpoint)+gridy		
					local currenttime=self.owner.currenttime
					self:SetBillboardAppearance(x,y,gx,gy,currenttime)
				else
					self.billboardgrid[x][y]:Hide()
				end
				
				--self.billboardgrid[x][y]:SetAllPoints(UIParent)
				-- if(x==1 and y==1)then
					-- TRAININGGROUNDS_FrameVisibilityCheck(self.billboardgrid[x][y])
				-- end
				
			end
		end
	 end
	
	
	
	function class:Cleanup()
		super.Cleanup(self)
		for i=1,#self.allsubframes do
			self.allsubframes[i]:Cleanup()
		end
		for i=1,#self.allbillboards do
			self.allbillboards[i]:Cleanup()
		end		
	end
	

	
	function TRAININGGROUNDS_SECRET35_MatrixMultiplicationRotation(x,y,z,ax,ay,az)
		local x2=x
		local y2=y
		local z2=z

														
		x2,y2,z2=TRAININGGROUNDS_SECRET35_MatrixMultiplication(x2,y2,z2,
											math.cos(az),-math.sin(az),0,
											math.sin(az),math.cos(az),0,
											0,0,1)		
		x2,y2,z2=TRAININGGROUNDS_SECRET35_MatrixMultiplication(x2,y2,z2,
											1,0,0,
											0,math.cos(ax),-math.sin(ax),
											0,math.sin(ax),math.cos(ax))
		x2,y2,z2=TRAININGGROUNDS_SECRET35_MatrixMultiplication(x2,y2,z2,
											math.cos(ay),0,math.sin(ay),
											0,1,0,
											-math.sin(ay),0,math.cos(ay))
		return x2,y2,z2
	end
	
	function TRAININGGROUNDS_SECRET35_MatrixMultiplication(x,y,z,a00,a10,a20,a01,a11,a21,a02,a12,a22)
		local x2,y2,z2
		x2=x*a00+y*a10+z*a20
		y2=x*a01+y*a11+z*a21
		z2=x*a02+y*a12+z*a22
		return x2,y2,z2
	end
	
	
	function TRAININGGROUNDS_SECRET35_PerspectiveTransform(x,y,z,far)
		-- Z is depth axis.  (TODO: this is still inconsistent with a few things)
		if(z<-far/2)then z=-far/2 end
		local x2,y2
		x2=x*far/(z+far)
		y2=y*far/(z+far)		
		--print(x2,y2)
		return x2,y2
	end
end
